"""
.. WARNING:: Except for :exc:`CoolqliteError` and :exc:`CoolqliteExceptionGroup`,
   these are unstable and may change.
"""

from coolqlite._common import (
    CoolqliteError,
    CoolqliteExceptionGroup,
    MissingNamedToSqlFnError,
    MissingTypedToSqlFnError,
    MissingToSqlFnError,
    ToSqlFnSetupError,
    ToSqlTypeError,
    FromSqlTypeError,
)
from coolqlite.transactions import (
    IncorrectTransactionControlError,
)

from coolqlite.query import (
    ParamMissingToSqlFnError,
    ToSqlFnFailure,
    QueryBuilderErrorGroup,
)

from coolqlite._results import ResultCountError

__all__ = [
    "CoolqliteError",
    "CoolqliteExceptionGroup",
    "MissingNamedToSqlFnError",
    "MissingTypedToSqlFnError",
    "MissingToSqlFnError",
    "ToSqlFnSetupError",
    "ToSqlTypeError",
    "FromSqlTypeError",
    "IncorrectTransactionControlError",
    "ParamMissingToSqlFnError",
    "ToSqlFnFailure",
    "QueryBuilderErrorGroup",
    "ResultCountError",
]
