"""A cool sqlite library."""

from coolqlite._connection import (
    Connection,
    Row,
    ToSqlFnRegistry,
    connect,
)
from coolqlite._results import Results, as_results, QueryResults, first_col
from coolqlite._common import RowFactory, SqliteType
from coolqlite.errors import (
    CoolqliteError,
    CoolqliteExceptionGroup,
    ResultCountError,
)


__all__ = [
    "connect",
    "Connection",
    "Results",
    "as_results",
    "QueryResults",
    "first_col",
    "Row",
    "CoolqliteError",
    "CoolqliteExceptionGroup",
    "ToSqlFnRegistry",
    "RowFactory",
    "SqliteType",
    "ResultCountError",
]
